// Small util to justify text with possible non-trivial unicode widths
// and terminal escape codes.

use unicode_width::UnicodeWidthStr;

pub(super) fn justify(input: &str, n: usize, fill: &str, dir: char) -> String {
    debug_assert!(terminal_width(fill) == 1);
    let w = terminal_width(input);
    if n > w {
        let d = n - w;
        match dir {
            '>' => format!("{}{input}", fill.repeat(d)),
            '<' => format!("{input}{}", fill.repeat(d)),
            '^' => {
                let (d, r) = (d / 2, d % 2);
                format!("{}{input}{}", fill.repeat(d), fill.repeat(d + r))
            }
            _ => panic!("Invalid text justification directive: {dir:?}."),
        }
    } else {
        input.into()
    }
}

// String width, not counting terminal escape codes, but taking utf8 widths into account.
pub(super) fn terminal_width(input: &str) -> usize {
    let stripped = strip_ansi_escapes::strip(input.as_bytes());
    let stripped = String::from_utf8(stripped).expect("Stripping escapes lead to invalid utf8?");
    UnicodeWidthStr::width(stripped.as_str())
}

#[cfg(test)]
mod tests {
    use colored::Colorize;

    use super::*;

    #[test]
    fn justify_strings() {
        let inputs = [
            format!(
                "{} {} {} {}",
                "✔".green(),
                "A".italic(),
                "🗙".red(),
                "B".bold()
            ),
            "V A X B".into(), // Should be the same.
        ];
        for input in inputs {
            assert_eq!(terminal_width(&input), 7);
            assert_eq!(justify(&input, 7, "-", '>'), input);
            assert_eq!(justify(&input, 8, "-", '>'), format!("-{input}"));
            assert_eq!(justify(&input, 8, "-", '<'), format!("{input}-"));
            assert_eq!(justify(&input, 8, "-", '^'), format!("{input}-"));
            assert_eq!(justify(&input, 9, "-", '>'), format!("--{input}"));
            assert_eq!(justify(&input, 9, "-", '<'), format!("{input}--"));
            assert_eq!(justify(&input, 9, "-", '^'), format!("-{input}-"));
        }
    }
}
