// Specify the content of the config file,
// and its representation within the program.
// Use TOML format for the file.
//
// The "raw" config data matches the content of the file.
// Once parsed, it is checked for consistency
// and transformed into the actual, useful config data.

use std::path::PathBuf;

use serde::Serialize;

use crate::{
    gene_tree::MeanNbBases,
    interner::SpeciesSymbol,
    optim::{bfgs::Config as BfgsConfig, gd::Config as GdConfig},
    BranchLength, Parameters,
};

pub mod check;
pub(crate) mod defaults;
mod deserialize;

// The final value handed out to user.
pub struct Config {
    // Path to the gene trees data file.
    // If relative: relative to the config file.
    pub trees: PathBuf,

    // Names of the species of interest.
    pub taxa: Taxa,

    // Parametrize the filters.
    pub filters: Filters,

    // If set, when the internal triplet branch length "d"
    // is less or equal to this value,
    // consider that its topology is not resolved enough
    // to exclude discordant scenarios.
    // In this situation,
    // every scenario contributes to the likelihood
    // instead of only the concordant ones.
    // The possible internal branch discordance
    // between 'actual' and 'expected' length
    // is neglected because the the actual length is small.
    // For this reason, *enforce* that it be small.
    // The value is given in *mutation* units, so branch length × sequence length.
    pub unresolved_length: Option<MeanNbBases>,

    // Parametrize the search.
    pub search: Search,
}
const MAX_UNRESOLVED_LENGTH: f64 = 0.5;

#[derive(Debug, Serialize)]
pub struct Filters {
    // When raised, filter out trees if some species in 'other'
    // branche between LCA(outgroup) and the root,
    // because they may should have been considered 'outgroup' instead.
    pub triplet_other_monophyly: bool,

    // Let `mt` be the mean branch length for gene tree triplet species,
    // and `mr` ("rest") be the mean branch length for outgroup/other species.
    // The "shape" of the gene tree is the ratio `q = mt/mr`.
    // The "global shape" or "mean shape" of all gene trees is `Q = smt/smr`,
    // where `smt` is the sum of all `mt` and `smr` of all `mr`
    // for all trees *with accepted topology*.
    // If set, this parameter reject trees whose shape differ from the mean shape
    // by more than a factor, say `2`, or anything specified with this parameter.
    // The motivation is that the hypothesis that mutation rate is constant
    // accross the tree(s) is too audacious when, eg. `q / Q ⩾ max_clock_ratio`.
    // Values need to be greater than 1
    // because they quantify "absolute" ratio or "imbalance": `max(q/Q, Q/q)`.
    pub max_clock_ratio: Option<BranchLength>,
}

// Contains only non-duplicated interned strings.
// Every section contains distinct species.
pub struct Taxa {
    pub triplet: SpeciesTriplet,
    pub(crate) outgroup: Vec<SpeciesSymbol>,
    pub(crate) other: Vec<SpeciesSymbol>,
}

// The triplet is importantly *oriented*: ((A, B), C).
// Accept several input forms provided they are unambiguous:
//    triplet = [["U", "V"], "W"] # A: U, B: V, C: W
//    triplet = ["U", ["V", "W"]] # A: V, B: W, C: U
// Or, shorter without special chars in species names:
//    triplet = "((U, V), W)"
//    triplet = "(U, (V, W))"
pub struct SpeciesTriplet {
    pub a: SpeciesSymbol,
    pub b: SpeciesSymbol,
    pub c: SpeciesSymbol,
}

#[derive(Debug, Serialize)]
pub struct Search {
    // Starting point in the search space.
    pub init_parms: Parameters<f64>,

    // Dataset reduction factor
    // when searching for non-degenerated starting point.
    pub(crate) init_data_reduction_factor: f64, // Above 1.

    // Gradient search configuration
    // when learning from a dataset reduction.
    pub(crate) init_descent: GdConfig,

    // Main learning phase configuration.
    pub(crate) bfgs: BfgsConfig,

    // Possible final gradient descent.
    pub(crate) final_descent: Option<GdConfig>,
}
