// Verify raw config consistency as we construct actual config from it.
// Take this opportunity to intern all species names and replace them with their symbols.

use std::{
    cmp::Ordering,
    collections::HashSet,
    path::{Path, PathBuf},
};

use arrayvec::ArrayVec;
use regex::Regex;
use snafu::{ensure, Snafu};

use crate::{
    config::{
        defaults,
        deserialize::{self as raw, NodeInput},
        Search, SpeciesTriplet, Taxa, MAX_UNRESOLVED_LENGTH,
    },
    interner::{Interner, SpeciesSymbol},
    io::{self, read_file},
    model::parameters::{GeneFlowTimes, MAX_N_GF_TIMES},
    optim::{
        self, bfgs::Config as BfgsConfig, gd::Config as GdConfig,
        wolfe_search::Config as WolfeConfig, SlopeTrackingConfig,
    },
    Config, Filters, Parameters,
};

macro_rules! err {
    ($fmt:tt) => {
        ConfigErr { mess: format!$fmt }
    };
}

impl Config {
    // Read raw config from file, then check it and convert into a valid config value.
    pub fn from_file(path: &Path, interner: &mut Interner) -> Result<Self, Error> {
        // Parse raw TOML file.
        let input = read_file(path)?;
        let raw = raw::Config::parse(&input)?;

        if let Some(ul) = raw.unresolved_length {
            ensure!(
                ul >= 0.,
                err!(
                    ("The minimum internal branch length for tree considered resolved \
                      cannot be negative. Received 'unresolved_length = {ul}'.")
                )
            );
            ensure!(
                ul <= MAX_UNRESOLVED_LENGTH,
                err!(
                    ("The minimum internal branch length for tree considered resolved \
                      is assumed to be small, \
                      so the program forbids that it be superior to {MAX_UNRESOLVED_LENGTH}. \
                      Received 'unresolved_length = {ul}'.")
                )
            );
        }

        let search = (&raw).try_into()?;
        let taxa = raw.taxa._try_into(interner)?;

        // Resolve relative paths relatively to the config file.
        let mut trees = raw.taxa.trees;
        if !trees.is_absolute() {
            let mut path = path
                .parent()
                .unwrap_or_else(|| {
                    panic!(
                        "Config file has been read but its path has no parent: {}",
                        path.display()
                    )
                })
                .to_owned();
            path.push(trees);
            trees = io::canonicalize(&path)?;
        }

        // Most checks implemented within `TryFrom` trait.
        Ok(Config {
            search,
            taxa,
            trees,
            unresolved_length: raw.unresolved_length,
            filters: if let Some(ref raw) = raw.filters {
                raw.try_into()?
            } else {
                defaults::filters()
            },
        })
    }

    // The union of all sections constitutes the set of "relevant" species.
    pub fn relevant_species(&self) -> impl Iterator<Item = SpeciesSymbol> + '_ {
        let Taxa { triplet, outgroup, other } = &self.taxa;
        triplet
            .iter()
            .chain(outgroup.iter().copied())
            .chain(other.iter().copied())
    }
}

//--------------------------------------------------------------------------------------------------
// Check filter parameters.
impl TryFrom<&raw::Filters> for Filters {
    type Error = Error;

    fn try_from(raw: &raw::Filters) -> Result<Self, Self::Error> {
        if let Some(mcr) = raw.max_clock_ratio {
            ensure!(
                mcr >= 1.,
                err!(
                    ("The maximum branch length ratio between triplet and outgroup \
                      cannot be lower than 1. Received 'max_clock_ratio = {mcr}'.")
                )
            );
        }
        Ok(Filters {
            max_clock_ratio: raw.max_clock_ratio,
            triplet_other_monophyly: raw
                .triplet_other_monophyly
                .unwrap_or(defaults::filters().triplet_other_monophyly),
        })
    }
}

//--------------------------------------------------------------------------------------------------
// Check search configuration.

impl TryFrom<&'_ raw::Config> for Search {
    type Error = Error;
    fn try_from(raw: &'_ raw::Config) -> Result<Self, Error> {
        let init_data_reduction_factor = raw.search.init_data_reduction_factor;
        if init_data_reduction_factor <= 1.0 {
            err!(
                ("The data reduction factor to search for non-degenerated starting point \
                  must be greater than 1. Received: {init_data_reduction_factor}.")
            );
        }
        Ok(Self {
            init_parms: Parameters::try_from(&raw.init, &raw.gf_times)?,
            init_data_reduction_factor,
            init_descent: (&raw.search.init_descent).try_into()?,
            bfgs: (&raw.search.bfgs).try_into()?,
            final_descent: raw
                .search
                .final_descent
                .as_ref()
                .map(TryInto::try_into)
                .transpose()?,
        })
    }
}

impl TryFrom<&'_ raw::GdConfig> for GdConfig {
    type Error = Error;
    fn try_from(raw: &raw::GdConfig) -> Result<Self, Self::Error> {
        let &raw::GdConfig { max_iter, learning_rate, ref slope_tracking } = raw;
        Ok(Self {
            max_iter,
            step_size: learning_rate,
            slope_tracking: slope_tracking
                .as_ref()
                .map(|sl| SlopeTrackingConfig::new(sl.sample_size, sl.threshold, sl.grain))
                .transpose()?,
        })
    }
}

impl TryFrom<&'_ raw::BfgsConfig> for BfgsConfig {
    type Error = Error;
    fn try_from(raw: &raw::BfgsConfig) -> Result<Self, Self::Error> {
        let &raw::BfgsConfig {
            max_iter,
            ref wolfe_search,
            step_size_threshold,
            ref slope_tracking,
            ref main_trace_path,
            ref linsearch_trace_path,
        } = raw;
        ensure!(
            0. <= step_size_threshold,
            err!(
                ("The threshold for defining small search steps \
                  must be null or positive. Received: {step_size_threshold}.")
            )
        );
        let check_path = |path: &Option<PathBuf>| -> Result<_, Self::Error> {
            let Some(path) = path else {
                return Ok(None);
            };
            let path = io::canonicalize(path)?;
            if path.is_file() {
                eprintln!("Will override {:?}.", path.to_string_lossy());
            };
            if let Some(parent) = path.parent() {
                ensure!(parent.exists(), NoSuchFolderErr { path });
            }
            Ok(Some(path))
        };
        Ok(Self {
            max_iter,
            slope_tracking: slope_tracking
                .as_ref()
                .map(|sl| SlopeTrackingConfig::new(sl.sample_size, sl.threshold, sl.grain))
                .transpose()?,
            wolfe: wolfe_search.try_into()?,
            small_step: step_size_threshold,
            // TODO: expose.
            main_trace_path: check_path(main_trace_path)?,
            linsearch_trace_path: check_path(linsearch_trace_path)?,
        })
    }
}

impl TryFrom<&raw::WolfeSearchConfig> for WolfeConfig {
    type Error = Error;
    fn try_from(raw: &raw::WolfeSearchConfig) -> Result<Self, Self::Error> {
        let &raw::WolfeSearchConfig {
            c1,
            c2,
            init_step_size,
            step_decrease_factor,
            step_increase_factor,
            flat_gradient,
            bisection_threshold,
        } = raw;
        for (c, which, name) in [(c1, "upper", "c1"), (c2, "lower", "c2")] {
            ensure!(
                0. < c && c < 1.,
                err!(
                    ("The Wolfe constant for checking {which} bound \
                      must stand between 0 and 1 (excluded). \
                      Received {name}={c1}.")
                )
            );
        }
        ensure!(
            c1 < c2,
            err!(
                ("The Wolfe constant c1 and c2 must verify c1 < c2. \
                  Received: c1 = {c1} ⩾ {c2} = c2.")
            )
        );
        ensure!(
            init_step_size > 0.,
            err!(
                ("Initial step size for Wolfe search must be positive. \
                  Received {init_step_size}.")
            )
        );
        ensure!(
            0. < step_decrease_factor && step_decrease_factor < 1.,
            err!(
                ("Decrease factor for Wolfe search must stand between 0 and 1 (excluded). \
                  Received: {step_decrease_factor}.")
            )
        );
        ensure!(
            1. < step_increase_factor,
            err!(
                ("Increase factor for Wolfe search must stand above 1. \
                  Received: {step_increase_factor}.")
            )
        );
        ensure!(
            0. <= flat_gradient,
            err!(
                ("Flat gradient threshold for Wolfe search must be positive or null. \
                  Received: {flat_gradient}.")
            )
        );
        ensure!(
            (0. ..1.).contains(&bisection_threshold),
            err!(
                ("The zoom bisection threshold for Wolfe search must stand between 0 and 1/2. \
                  Received: {bisection_threshold}.")
            )
        );
        Ok(Self {
            c1,
            c2,
            init_step: init_step_size,
            step_decrease: step_decrease_factor,
            step_increase: step_increase_factor,
            flat_gradient,
            bisection_threshold,
        })
    }
}

//--------------------------------------------------------------------------------------------------
// Check initial parameters.

impl Parameters<f64> {
    fn try_from(
        init: &raw::InitialParameters,
        raw_gft: &raw::GeneFlowTimes,
    ) -> Result<Self, Error> {
        use Ordering as O;
        let &raw::InitialParameters {
            theta,
            tau_1,
            tau_2,
            p_ab,
            p_ac,
            p_bc,
            p_ancient_gf,
        } = init;

        macro_rules! check_positive {
            ($($name:ident),+) => {$({
                let value = init.$name;
                ensure!(value >= 0.,
                    err!((
                        "Model parameter '{}' must be positive, not {value}.",
                        stringify!($name)
                    ))
                )
            })+};
        }
        macro_rules! check_prob {
            ($($name:ident),+) => {$({
                let value = init.$name;
                ensure!((0. ..=1.).contains(&value),
                    err!((
                        "Model parameter '{}' must be a probability (between 0 and 1), \
                         not {value}.",
                        stringify!($name)
                    ))
                )
            })+};
        }

        check_positive!(theta, tau_1, tau_2);
        check_prob!(p_ab, p_ac, p_bc, p_ancient_gf);

        match tau_2.total_cmp(&tau_1) {
            O::Less => {
                return err!(
                    ("The second coalescence time must be older than the first: \
                      maybe tau_1 ({tau_1}) and tau_2 ({tau_2}) were meant to be reversed?")
                )
                .fail()
            }
            O::Equal => {
                return err!(
                    ("The two coalescence times cannot be identical: \
                      here tau_1 == tau_2 == {tau_1}.")
                )
                .fail()
            }
            O::Greater => {}
        };

        let s = p_ab + p_bc + p_ac;
        ensure!(
            s <= 1.0,
            err!(
                ("The sum of gene flow transfer probabilities must not be larger than 1. \
                  p_ab + p_bc + p_ac = {p_ab} + {p_bc} + {p_ac} = {s} > 1.")
            )
        );

        Ok(Parameters {
            theta,
            tau_1,
            tau_2,
            p_ab,
            p_ac,
            p_bc,
            p_ancient: p_ancient_gf,
            gf_times: raw_gft.try_into()?,
        })
    }
}

//--------------------------------------------------------------------------------------------------
// Check gene flow times.

impl TryFrom<&raw::GeneFlowTimes> for GeneFlowTimes<f64> {
    type Error = Error;
    fn try_from(raw: &raw::GeneFlowTimes) -> Result<Self, Self::Error> {
        if raw.is_empty() {
            err!(("No gene flow time specified."));
        }

        if raw.len() > MAX_N_GF_TIMES {
            err!((
                "Aphid does not support more than {MAX_N_GF_TIMES:?} gene flow times, \
                 but {} times were provided.",
                raw.len()
            ));
        }

        // Check values while sorting them.
        let mut gf_times = Self::new();
        for &time in raw {
            if time.is_infinite() || time.is_nan() || time < 0. {
                err!(("Invalid gene flow time: {time}."));
            }
            if time > 1.0 {
                err!(
                    ("Gene flow time cannot yet be specified earlier \
                      than latest species divergence, but {time} is greater than 1.")
                );
            }
            match gf_times
                .0
                .binary_search_by(|&t| t.total_cmp(&time).reverse())
            {
                Ok(_) => return err!(("Gene flow time '{time}' specified twice.")).fail(),
                Err(i) => gf_times.0.insert(i, time),
            }
        }
        Ok(gf_times)
    }
}

//--------------------------------------------------------------------------------------------------
impl raw::Taxa {
    fn _try_into(&self, interner: &mut Interner) -> Result<Taxa, Error> {
        // Check taxa consistency.
        let Self { triplet: raw_triplet, outgroup, other, .. } = self;
        let triplet: [&str; 3] = raw_triplet
            .try_into()
            .map_err(|mess| Error::Config { mess })?;
        if outgroup.is_empty() {
            err!(("No outgroup species specified."));
        }

        // No duplicates should be found among the taxa explicitly given.
        let mut relevant_species = HashSet::new();
        let mut new_species = move |name, section| -> Result<_, Error> {
            ensure!(
                !relevant_species.contains(&name),
                err!(
                    ("Species {name:?} found in '{section}' section \
                      is given twice within the [taxa] table.")
                )
            );
            let symbol = interner.get_or_intern(name);
            relevant_species.insert(name);
            // Also record into the interner.
            Ok(symbol)
        };

        let [a, b, c] = triplet.map(|s| new_species(s, "triplet"));
        Ok(Taxa {
            triplet: SpeciesTriplet { a: a?, b: b?, c: c? },
            outgroup: outgroup
                .iter()
                .map(|s| new_species(s, "outgroup"))
                .collect::<Result<Vec<_>, _>>()?,
            other: other
                .iter()
                .map(|s| new_species(s, "other"))
                .collect::<Result<Vec<_>, _>>()?,
        })
    }
}

//--------------------------------------------------------------------------------------------------
// Check triplet and extract strings in canonical order.

impl<'t> TryFrom<&'t raw::Triplet> for [&'t str; 3] {
    type Error = String;

    // Work towards canonicalized ((a, b), c),
    // (x, y) refer to input nodes still undetermined.
    #[allow(clippy::many_single_char_names)]
    fn try_from(input: &'t raw::Triplet) -> Result<Self, Self::Error> {
        use raw::Triplet as T;
        use NodeInput as N;
        macro_rules! err {
            ($($message:tt)+) => {{
                return Err(format!$($message)+);
            }};
        }
        match input {
            T::Arrays(ars) => {
                let original = format!("{ars:?}");
                // Special-case one useful error.
                if ars.len() == 3 {
                    let mut species = ArrayVec::<&str, 3>::new();
                    for i in ars {
                        match i {
                            N::Leaf(sp) => species.push(sp),
                            N::Internal(_) => break,
                        }
                    }
                    if species.len() == 3 {
                        let [a, b, c] = species.into_inner().unwrap();
                        err!(
                            ("Ambiguous triplet:\n  {original}\nSpecify either\n  \
                              [[{a:?}, {b:?}], {c:?}]\nor\n   [{a:?}, [{b:?}, {c:?}]]")
                        );
                    }
                }
                // Fallback error for other invalid input types.
                macro_rules! invalid { () => {{
                    err!(
                        ("Invalid triplet specification. Expected 3 nested nodes under the form \
                         [[\"A\", \"B\"], \"C\"] or [\"C\", [\"A\", \"B\"]], got {original}.")
                    );
                }};}
                if ars.len() != 2 {
                    invalid!()
                }
                let mut it = ars.iter();
                let (x, y) = (it.next().unwrap(), it.next().unwrap());
                let ((N::Internal(ab), N::Leaf(c)) | (N::Leaf(c), N::Internal(ab))) = (x, y) else {
                    invalid!()
                };
                if ab.len() != 2 {
                    invalid!()
                }
                let mut it = ab.iter();
                let (a, b) = (it.next().unwrap(), it.next().unwrap());
                Ok([a, b, c])
            }
            T::Parenthesized(input) => {
                // A small tree with 3 expected nodes can be just regular-parsed.
                // Capture groups for trimmed leaf nodes.
                let specials = ",()"; // Forbidden in species names then.
                let mut it = "abc"
                    .chars()
                    .map(|x| format!(r"\s*(?<{x}>[^{specials}]+?)\s*"));
                let (a, b, c) = (it.next().unwrap(), it.next().unwrap(), it.next().unwrap());
                // Match internal node.
                let ab = format!(r"\s*\({a},{b}\)\s*");
                // There are two possible forms.
                let [alt1, alt2] = [format!("{ab},{c}"), format!("{c},{ab}")]
                    .map(|p| Regex::new(&format!(r"\s*\({p}\)\s*")).unwrap());
                // Match either for the whole triplet.
                let caps = if let Some(caps) = alt1.captures(input) {
                    caps
                } else if let Some(caps) = alt2.captures(input) {
                    caps
                } else {
                    err!(
                        ("Invalid triplet specification. Expected 3 nested nodes under the form \
                         \"((A, B), C)\" or \"(C, (A, B))\", got {input:?}.")
                    )
                };
                let extr = |x| caps.name(x).unwrap().as_str();
                Ok([extr("a"), extr("b"), extr("c")])
            }
        }
    }
}

impl SpeciesTriplet {
    pub fn as_array(&self) -> [SpeciesSymbol; 3] {
        [self.a, self.b, self.c]
    }

    pub fn iter(&self) -> impl Iterator<Item = SpeciesSymbol> {
        self.as_array().into_iter()
    }
}

//==================================================================================================
// Errors.

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(transparent)]
    Io { source: io::Error },
    #[snafu(transparent)]
    Parse { source: raw::Error },
    #[snafu(display("Configuration error: {mess}"))]
    Config { mess: String },
    #[snafu(transparent)]
    Optim { source: optim::Error },
    #[snafu(display("Could not find folder: {:?}", path.display()))]
    NoSuchFolder { path: PathBuf },
}

//==================================================================================================
// Test.

#[cfg(test)]
mod tests {
    use serde::Deserialize;

    use super::*;

    #[test]
    #[allow(clippy::too_many_lines)]
    fn triplet_parsing() {
        #[derive(Deserialize)]
        struct Single {
            raw: raw::Triplet,
        }
        let success = |input, expected| {
            println!("Checking success for: {input}.");
            let raw: Single = toml::from_str(&format!("raw = {input}")).unwrap();
            let actual: [&str; 3] = (&raw.raw).try_into().unwrap();
            assert_eq!(expected, actual);
        };
        success(r#""((U, V), W)""#, ["U", "V", "W"]);
        success(r#""(U, (V, W))""#, ["V", "W", "U"]);
        success(
            r#"" ( ( with space , pretty much ) , everywhere ) ""#,
            ["with space", "pretty much", "everywhere"],
        );
        success(r#"[["U", "V"], "W"]"#, ["U", "V", "W"]);
        success(r#"["U", ["V", "W"]]"#, ["V", "W", "U"]);

        let failure = |input, expected: String| {
            println!("Checking failure for: {input}.");
            let raw = match toml::from_str::<Single>(&format!("raw = {input}")) {
                Ok(r) => r,
                Err(e) => {
                    let actual = format!("{e}");
                    assert_eq!(expected, actual, "Invalid error message.");
                    return;
                }
            };
            let res: [&str; 3] = match (&raw.raw).try_into() {
                Ok(r) => r,
                Err(e) => {
                    let actual = e.to_string();
                    assert_eq!(expected, actual, "Invalid error message.");
                    return;
                }
            };
            panic!("Unexpected success with\n  {input:?}\nObtained:\n  {res:?}");
        };
        failure(
            r#""(U, V, W)""#,
            concat!(
                r#"Invalid triplet specification. Expected 3 nested nodes under the form "#,
                r#""((A, B), C)" or "(C, (A, B))", got "(U, V, W)"."#
            )
            .into(),
        );
        failure(
            r#""(U, V, W, X)""#,
            concat!(
                r#"Invalid triplet specification. Expected 3 nested nodes under the form "#,
                r#""((A, B), C)" or "(C, (A, B))", got "(U, V, W, X)"."#,
            )
            .into(),
        );
        failure(
            r#""((U, V, W), X)""#,
            concat!(
                r#"Invalid triplet specification. Expected 3 nested nodes under the form "#,
                r#""((A, B), C)" or "(C, (A, B))", got "((U, V, W), X)"."#,
            )
            .into(),
        );
        failure(
            r#""((U, V), W, X)""#,
            concat!(
                r#"Invalid triplet specification. Expected 3 nested nodes under the form "#,
                r#""((A, B), C)" or "(C, (A, B))", got "((U, V), W, X)"."#,
            )
            .into(),
        );
        failure(
            r#""((U, V), W, X)""#,
            concat!(
                r#"Invalid triplet specification. Expected 3 nested nodes under the form "#,
                r#""((A, B), C)" or "(C, (A, B))", got "((U, V), W, X)"."#,
            )
            .into(),
        );
        // Special-cased error message.
        failure(
            r#"["U", "V", "W"]"#,
            concat!(
                r#"Ambiguous triplet:\n  ["U", "V", "W"]\n"#,
                r#"Specify either\n  [["U", "V"], "W"]\n"#,
                r#"or\n   ["U", ["V", "W"]]"#,
            )
            .replace(r"\n", "\n"),
        );
        failure(
            r#"["U", "V", "W", "X"]"#,
            concat!(
                r#"Invalid triplet specification. "#,
                r#"Expected 3 nested nodes under the form "#,
                r#"[["A", "B"], "C"] or ["C", ["A", "B"]], "#,
                r#"got ["U", "V", "W", "X"]."#
            )
            .into(),
        );
        failure(
            r#"[["U", "V"], "W", "X"]"#,
            concat!(
                r#"Invalid triplet specification. "#,
                r#"Expected 3 nested nodes under the form "#,
                r#"[["A", "B"], "C"] or ["C", ["A", "B"]], "#,
                r#"got [["U", "V"], "W", "X"]."#
            )
            .into(),
        );
        failure(
            r#"[["U", "V"], ["W", "X"]]"#,
            concat!(
                r#"Invalid triplet specification. "#,
                r#"Expected 3 nested nodes under the form "#,
                r#"[["A", "B"], "C"] or ["C", ["A", "B"]], "#,
                r#"got [["U", "V"], ["W", "X"]]."#
            )
            .into(),
        );
        failure(
            r"[]",
            concat!(
                r#"Invalid triplet specification. "#,
                r#"Expected 3 nested nodes under the form "#,
                r#"[["A", "B"], "C"] or ["C", ["A", "B"]], "#,
                r#"got []."#
            )
            .into(),
        );
        // Not a super error.. but go with it for now.
        failure(
            r#"[["U", ["V", "W"]]]"#,
            concat!(
                r#"TOML parse error at line 1, column 7\n"#,
                r#"  |\n"#,
                r#"1 | raw = [["U", ["V", "W"]]]\n"#,
                r#"  |       ^^^^^^^^^^^^^^^^^^^\n"#,
                r#"data did not match any variant of untagged enum Triplet\n"#,
            )
            .replace(r"\n", "\n"),
        );
    }
}
