// Decisions for anything not user-provided.

use crate::{
    config::deserialize::{
        BfgsConfig, GdConfig, InitialParameters, Search, SlopeTrackingConfig, WolfeSearchConfig,
    },
    Filters,
};

pub(crate) fn initial_parameters() -> InitialParameters {
    InitialParameters {
        theta: 0.005,
        tau_1: 0.002,
        tau_2: 0.005,
        p_ab: 0.3,
        p_ac: 0.3,
        p_bc: 0.3,
        p_ancient_gf: 0.7,
    }
}

pub(crate) fn gf_times() -> Vec<f64> {
    vec![1.]
}

pub(crate) fn filters() -> Filters {
    Filters {
        max_clock_ratio: None,
        triplet_other_monophyly: false,
    }
}

pub(crate) fn search() -> Search {
    Search {
        init_data_reduction_factor: init_data_reduction_factor(),
        init_descent: init_descent(),
        bfgs: bfgs(),
        final_descent: None,
    }
}

pub(crate) fn init_data_reduction_factor() -> f64 {
    2.0
}

pub(crate) fn init_descent() -> GdConfig {
    GdConfig {
        max_iter: 1_000,
        learning_rate: 1e-2,
        slope_tracking: Some(SlopeTrackingConfig { sample_size: 10, threshold: 1e-1, grain: 5 }),
    }
}

pub(crate) fn bfgs() -> BfgsConfig {
    BfgsConfig {
        max_iter: 1_000,
        wolfe_search: wolfe_search(),
        step_size_threshold: 1e-9,
        slope_tracking: Some(SlopeTrackingConfig { sample_size: 20, threshold: 1e-3, grain: 5 }),
        main_trace_path: None,
        linsearch_trace_path: None,
    }
}

pub(crate) fn wolfe_search() -> WolfeSearchConfig {
    // Taken from Nocedal and Wright 2006.
    let c1 = 1e-4;
    let c2 = 0.1;
    WolfeSearchConfig {
        c1,
        c2,
        init_step_size: 1.0,
        step_decrease_factor: c2,
        step_increase_factor: 10., // (custom addition)
        flat_gradient: 1e-20,
        bisection_threshold: 1e-1,
    }
}

#[allow(dead_code)] // Not the default anymore.
pub(crate) fn final_descent() -> GdConfig {
    GdConfig {
        max_iter: 1_000,
        learning_rate: 1e-5,
        slope_tracking: Some(SlopeTrackingConfig { sample_size: 100, threshold: 1e-3, grain: 50 }),
    }
}
