// Deserialize raw config from file.
// This type is a faithful representation
// what the user wrote in the config file,
// and it then needs to be checked
// into the eventual config type handed to user.

use std::{array, fmt, path::PathBuf};

use serde::Deserialize;
use serde_with::{serde_as, FromInto};
use snafu::{ResultExt, Snafu};

use crate::config::defaults;

//--------------------------------------------------------------------------------------------------
// Global config structure and defaults.

pub(crate) type GeneFlowTimes = Vec<f64>;

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(crate) struct Config {
    pub(crate) filters: Option<Filters>,
    pub(crate) taxa: Taxa,
    #[serde(default = "defaults::gf_times")]
    pub(crate) gf_times: GeneFlowTimes,
    pub(crate) unresolved_length: Option<f64>,
    #[serde(default = "defaults::initial_parameters")]
    pub(crate) init: InitialParameters,
    #[serde(default = "defaults::search")]
    pub(crate) search: Search,
}

impl Config {
    // Parse config file successfully
    // or abort the program with a useful error.
    pub(crate) fn parse(input: &str) -> Result<Self, Error> {
        toml::from_str(input).context(DeserializeErr {})
    }
}

//--------------------------------------------------------------------------------------------------
// Filters configuration.

#[serde_as]
#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(crate) struct Filters {
    pub(crate) max_clock_ratio: Option<f64>,
    pub(crate) triplet_other_monophyly: Option<bool>,
}

//--------------------------------------------------------------------------------------------------
// Species data.

#[serde_as]
#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(crate) struct Taxa {
    // Where to find the gene trees.
    pub(crate) trees: PathBuf,

    pub(crate) triplet: Triplet,

    #[serde_as(as = "FromInto<ListOfStrings>")]
    pub(crate) outgroup: Vec<String>,

    #[serde_as(as = "FromInto<ListOfStrings>")]
    pub(crate) other: Vec<String>,
}

// Guard against invalid inputs.
#[derive(Deserialize)]
#[serde(untagged)]
pub(crate) enum Triplet {
    Arrays(Vec<NodeInput>),
    Parenthesized(String),
}

#[derive(Deserialize)]
#[serde(untagged)]
pub(crate) enum NodeInput {
    Leaf(String),
    Internal(Vec<String>),
}

impl fmt::Debug for Triplet {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        use Triplet as T;
        match self {
            T::Arrays(ars) => write!(f, "{ars:?}"),
            T::Parenthesized(s) => write!(f, "{s:?}"),
        }
    }
}

impl fmt::Debug for NodeInput {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        use NodeInput as N;
        match self {
            N::Leaf(s) => write!(f, "{s:?}"),
            N::Internal(vec) => write!(f, "{vec:?}"),
        }
    }
}

//--------------------------------------------------------------------------------------------------
// Statistical model parameters.

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(crate) struct InitialParameters {
    pub(crate) theta: f64,
    pub(crate) tau_1: f64,
    pub(crate) tau_2: f64,
    pub(crate) p_ab: f64,
    pub(crate) p_ac: f64,
    pub(crate) p_bc: f64,
    pub(crate) p_ancient_gf: f64,
}

//--------------------------------------------------------------------------------------------------
// Search configuration.

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(crate) struct Search {
    #[serde(default = "defaults::init_data_reduction_factor")]
    pub(crate) init_data_reduction_factor: f64,
    #[serde(default = "defaults::init_descent")]
    pub(crate) init_descent: GdConfig,
    #[serde(default = "defaults::bfgs")]
    pub(crate) bfgs: BfgsConfig,
    pub(crate) final_descent: Option<GdConfig>,
}

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(crate) struct BfgsConfig {
    pub(crate) max_iter: u64,
    #[serde(default = "defaults::wolfe_search")]
    pub(crate) wolfe_search: WolfeSearchConfig,
    pub(crate) step_size_threshold: f64,
    pub(crate) slope_tracking: Option<SlopeTrackingConfig>,
    pub(crate) main_trace_path: Option<PathBuf>,
    pub(crate) linsearch_trace_path: Option<PathBuf>,
}

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(crate) struct WolfeSearchConfig {
    pub(crate) c1: f64,
    pub(crate) c2: f64,
    pub(crate) init_step_size: f64,
    pub(crate) step_decrease_factor: f64,
    pub(crate) step_increase_factor: f64,
    pub(crate) flat_gradient: f64,
    pub(crate) bisection_threshold: f64,
}

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(crate) struct GdConfig {
    pub(crate) max_iter: u64,
    pub(crate) learning_rate: f64,
    pub(crate) slope_tracking: Option<SlopeTrackingConfig>,
}

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(crate) struct SlopeTrackingConfig {
    pub(crate) sample_size: usize,
    pub(crate) threshold: f64,
    pub(crate) grain: u64,
}

//==================================================================================================
// Parsing utils.

// Lists of strings can either be specified as regular arrays like in:
//    ["a", "b", "c", "d"]
// Or as whitespace-separated strings, like in:
//     "a b c d"
// The latter is shorter,
// but less flexible as it does not allow whitespace in strings.
#[derive(Deserialize)]
#[serde(untagged)]
enum ListOfStrings {
    Regular(Vec<String>),
    Compact(String),
}

impl From<ListOfStrings> for Vec<String> {
    fn from(lof: ListOfStrings) -> Self {
        use ListOfStrings as L;
        match lof {
            L::Regular(vec) => vec,
            L::Compact(input) => input.split_whitespace().map(ToOwned::to_owned).collect(),
        }
    }
}

impl<const N: usize> TryFrom<ListOfStrings> for [String; N] {
    type Error = String;

    fn try_from(lof: ListOfStrings) -> Result<Self, Self::Error> {
        use ListOfStrings as L;
        let vec = match lof {
            L::Regular(vec) => vec,
            L::Compact(input) => input.split_whitespace().map(ToOwned::to_owned).collect(),
        };
        if vec.len() == N {
            let mut it = vec.into_iter();
            Ok(array::from_fn(|_| it.next().unwrap()))
        } else {
            Err(format!(
                "Expected {N} items in list of strings, found {}.",
                vec.len()
            ))
        }
    }
}

//==================================================================================================
// Errors.

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(display(
        "Error while reading config file:\n{}",
        complete_deserialize_error(source)
    ))]
    Deserialize { source: toml::de::Error },
}

// Best-attempt to improve over https://github.com/serde-rs/serde/issues/773.
fn complete_deserialize_error(e: &toml::de::Error) -> String {
    let mut message = format!("{e}");
    let pattern = "data did not match any variant of untagged enum ";
    if message.contains(pattern) {
        let name = message.trim().rsplit(' ').next().unwrap();
        let accepted_types = match name {
            // TODO: generate these with procedural macros?
            "ListOfStrings" => vec!["Array", "String"],
            _ => panic!("Error in source code: enum {name} has not been handled."),
        };
        message = message.replace(
            &format!("{pattern}{name}"),
            &format!(
                "invalid type: expected one of [{}]",
                accepted_types.join(", ")
            ),
        );
    };
    message
}
