// Reading from / writing to files on disk.

use std::{
    fs::File,
    io::{self, Read},
    path::{Path, PathBuf},
    string::FromUtf8Error,
};

use snafu::{ResultExt, Snafu};

// Read into owned data, emiting dedicated crate errors on failure.
pub(crate) fn read_file(path: &Path) -> Result<String, Error> {
    let tobuf = || -> PathBuf { path.into() };
    let mut file = File::open(path).context(IoErr { path: tobuf() })?;
    let mut buf = Vec::new();
    file.read_to_end(&mut buf)
        .context(IoErr { path: tobuf() })?;
    String::from_utf8(buf).context(Utf8Err { path: tobuf() })
}

// Wrap to integrate into aphid error system.
pub fn canonicalize(path: &Path) -> Result<PathBuf, Error> {
    path.canonicalize()
        .with_context(|_| CanonicalizeErr { path })
}

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(display("IO error while reading file {}:\n{source}", path.display()))]
    Io { path: PathBuf, source: io::Error },
    #[snafu(display("UTF-8 encoding error in file {}:\n{source}", path.display()))]
    Utf8 {
        path: PathBuf,
        source: FromUtf8Error,
    },
    #[snafu(display("Could not canonicalize path: {:?}:\n{source}", path.display()))]
    Canonicalize {
        source: std::io::Error,
        path: PathBuf,
    },
}
