// Use homebrew gradient descent and BFGS search
// to find scores values yielding maximum log-likelihood.
// Model log-likelihood as -F(X, P)
// with 'X' the 'data', constant throughout the optimisation,
// and 'P' the 'parameters' to optimize,
// and that we wish to derive F with respect to.

use snafu::{ensure, Snafu};
use tch::{Device, Tensor};

use crate::{
    config::Search,
    model::{
        likelihood::{data_tensors, ln_likelihood_tensors},
        parameters::GeneFlowTimes,
        scenarios::Scenario,
        scores::Scores,
    },
    optim::{Error as OptimError, Optim, OptimResult, OptimTensor},
    GeneTriplet, Parameters,
};

// Return both optimized inputs and output.
#[allow(clippy::similar_names)] // p_ab/p_ac is okay.
#[allow(clippy::too_many_lines)] // TODO: Fix once stabilized.
pub fn optimize_likelihood(
    // Receive in this format for better locality.
    triplets: &[GeneTriplet],
    start: &Parameters<f64>,
    search: &Search,
) -> Result<(Parameters<f64>, f64), Error> {
    // Tensors are small and control flow depends a lot on their values,
    // so roundtrips to a Gpu are not exactly interesting
    // in terms of performances.
    let device = Device::Cpu;
    let n_triplets = triplets.len();

    // Extract parameters tensors 'P', tracked for the gradient.
    // Some few "fixed parameters", part of X, can be retrieved within the scores.
    let (mut p, scores) = leaves_tensors(start, device);

    // Extract all remaining data 'X' from the triplets.
    let ngf = start.n_gf_times();
    let scenarios = Scenario::iter(ngf).collect::<Vec<_>>();
    let n_scenarios = scenarios.len();
    let x = data_tensors(triplets, n_scenarios, &scenarios, device);

    let mut n_eval = 0;
    let mut n_diff = 0;

    // Check starting point.
    let first_lnl = ln_likelihood_tensors(&x, &scores(&p)).to_double();
    if !first_lnl.is_finite() {
        // When there are a lot of data,
        // the starting point parameters 'P0' may lead
        // to numerically non-finite likelihood values.
        // If this happens, perform the optimization
        // on a smaller batch of the data first until we get finite likelihood.
        let sc: Scores<f64> = (&scores(&p)).into();
        eprintln!(
            "The chosen starting point yields non-finite log-likelihood ({first_lnl}) \
             on the whole dataset ({n_triplets} triplets):\n{sc}\n{start:?}",
        );

        // Numerical conversion shenanigans to divide dataset by a floating factor.
        let reduce = |n: usize| {
            let n: u64 = n as u64;
            #[allow(clippy::cast_precision_loss)]
            let n: f64 = n as f64;
            let r = n / search.init_data_reduction_factor;
            #[allow(clippy::cast_sign_loss, clippy::cast_possible_truncation)]
            let r: usize = r as usize;
            r
        };

        // Loop to search a starting point that works for all data.
        'p: loop {
            let mut n_samples = reduce(triplets.len());
            let (mut sub_triplets, mut sub_x);
            // Loop to search a data subset that works for the current starting point.
            'x: loop {
                ensure!(
                    n_samples > 0,
                    NonFiniteLikelihoodErr { lnl: first_lnl, parms: start.clone() }
                );
                eprintln!(
                    "-- Try obtaining finite log-likelihood \
                        with the {n_samples} first triplets."
                );
                sub_triplets = &triplets[0..n_samples];
                sub_x = data_tensors(sub_triplets, n_scenarios, &scenarios, device);
                let lnl = ln_likelihood_tensors(&sub_x, &scores(&p)).to_double();
                n_eval += 1;
                if !lnl.is_finite() {
                    eprintln!("---- Failure: obtained log-likelihood: {lnl}.");
                    // Reduce the data again.
                    n_samples = reduce(n_samples);
                    continue 'x;
                }
                // With this working (sub_X, P), perform a learning pass
                // to produce a (hopefully better) candidate P.
                eprintln!(
                    "---- Success: obtained log-likelihood: {lnl}.\n\
                     -- Learn on this subsample with simple gradient descent."
                );
                let f = |p: &Tensor| -ln_likelihood_tensors(&sub_x, &scores(p));
                match search.init_descent.minimize(f, &p, 0) {
                    Err(e) => {
                        eprintln!("---- Learning failed:\n{e}");
                        n_samples /= 2;
                        continue 'x;
                    }
                    Ok(opt) => {
                        n_eval += opt.n_eval();
                        n_diff += opt.n_diff();
                        p = opt.best_vars().copy();
                        let sc: Scores<f64> = (&scores(&p)).into();
                        let pm: Parameters<f64> = (&sc).into();
                        eprintln!(
                            "---- Success: parameters learned on the subsample:\n{sc}\n{pm:?}"
                        );
                        break 'x;
                    }
                }
            }
            eprintln!("-- Try with this new starting point.");
            let lnl = ln_likelihood_tensors(&x, &scores(&p)).to_double();
            n_eval += 1;
            if !lnl.is_finite() {
                eprintln!(
                    "---- Failure: obtained non-finite log-likelihood again ({lnl}). \
                     Try subsampling again from the new starting point."
                );
                continue 'p;
            }
            eprintln!(
                "---- Success: obtained finite log-likelihood on the whole dataset ({lnl}).\n\
                 -- Start learning from this new starting point."
            );
            break 'p;
        }
    }

    // Learn over the whole dataset.
    let f = |p: &Tensor| -ln_likelihood_tensors(&x, &scores(p));

    println!("-- BFGS learning.");
    let mut opt: Box<dyn OptimResult>;
    opt = Box::new(search.bfgs.minimize(f, &p, 0)?);
    let p = opt.best_vars();
    n_eval += opt.n_eval();
    n_diff += opt.n_diff();

    // Then simple gradient descent to refine.
    if let Some(gd) = &search.final_descent {
        println!("-- Refine with simple gradient descent.");
        opt = Box::new(gd.minimize(f, p, 0)?);
    }
    n_eval += opt.n_eval();
    n_diff += opt.n_diff();

    // Final value.
    let p = opt.best_vars().set_requires_grad(true);
    let loss = f(&p);
    loss.backward();
    let grad = p.grad();
    let scores = scores(&p);
    let grad = scores.with_grads(&grad);
    println!("-- Terminate heuristic after {n_eval} evaluations and {n_diff} differentiations.");
    println!("-- Final gradient:\n{grad:#}");

    let parms: Parameters<f64> = (&scores).into();
    let opt_lnl = -opt.best_loss();
    Ok((parms, opt_lnl))
}

// Construct all gradient-tracked parameters,
// gathered into a full 'P' tensor.
// Also construct a closure
// to extract them into a `Scores` value
// used as input to the likelihood formula.
// The Scores gather (aliased- = non-leaves-)gradient-tracked parameters
// and the untracked parameters,
// which are then conceptually part of the data 'X'.
fn leaves_tensors(
    start: &Parameters<f64>,
    dev: Device,
) -> (Tensor, impl Fn(&Tensor) -> Scores<Tensor>) {
    // Calculate free-ranging scores from the desired starting point.
    let s: Scores<_> = start.into();
    let Scores {
        theta,
        tau_1,
        delta_tau,
        gf,
        ac,
        bc,
        ancient,
        gf_times,
    } = s;

    // TODO: make it configurable which ones are tracked.
    let tracked =
        Tensor::from_slice(&[theta, tau_1, delta_tau, gf, ac, bc, ancient]).to_device(dev);
    // Scores `gf_times` are not tracked yet.

    (
        tracked,
        // Specify how to extract typed, mixed '(P, X)' scores from the pure 'P' tensor.
        move |p| {
            let mut it = (0..).map(|i| p.get(i));
            let mut next = || it.next().unwrap();

            Scores {
                // /!\ Order here must match order within the tensor: match it!
                theta: next(),
                tau_1: next(),
                delta_tau: next(),
                gf: next(),
                ac: next(),
                bc: next(),
                ancient: next(),
                gf_times: GeneFlowTimes(
                    gf_times
                        .0
                        .iter()
                        .map(|&t| Tensor::from(t).to_device(dev))
                        .collect(),
                ),
            }
        },
    )
}

//==================================================================================================
// Errors.

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(display(
        "The following initial parameters yielded non-finite log-likelihood ({lnl}) \
         even when (naively) subsampling the data:\n{parms:#?}"
    ))]
    NonFiniteLikelihood { lnl: f64, parms: Parameters<f64> },
    #[snafu(transparent)]
    Optim { source: OptimError },
}
