// Structure the parameters needed for calculating the likelihood.
// "Parameters" in the sense that they may be subject to optimisation.

use std::fmt::{self, Display};

use arrayvec::ArrayVec;
use serde::Serialize;
use tch::Tensor;

// The structure is generic among float and tensors
// to allow both likelihood implementations.
// NB: the required bounds to abstract over both in this program
// is very complicated: (f64 - Num, Num / f64, Num.exp(), etc.)
// and it needs to be repeated on every method implementation.
// As a consequence, either duplicate the code or use macros,
// and keep genericity just for factorizing over the parameters type.
pub trait Num: Sized {}
impl Num for f64 {}
impl Num for Tensor {}
impl Num for ValGrad {}

// Constraints to uphold:
//  - all positive
//  - tau_1 <= tau_2
//  - p_* <= 1
//  - p_ab + p_ac + p_bc <= 1
#[derive(Debug, Clone, Serialize)]
pub struct Parameters<F: Num> {
    // Population size.
    pub(crate) theta: F,
    // Divergence times.
    pub(crate) tau_1: F,
    pub(crate) tau_2: F,
    // Gene flow.
    pub(crate) p_ab: F,
    pub(crate) p_ac: F,
    pub(crate) p_bc: F,
    pub(crate) p_ancient: F, // Only the most ancient gf time gets a special probability.
    // Date of inferred gene flow events.
    pub(crate) gf_times: GeneFlowTimes<F>, // At least one value, sorted decreasing, all positives.
}

// Parameters are stack-allocated, so don't overuse possible GF times.
pub(crate) const MAX_N_GF_TIMES: usize = 3;
#[derive(Serialize, Debug, Clone, PartialEq, Eq)]
#[serde(transparent)]
pub struct GeneFlowTimes<F: Num>(pub(crate) ArrayVec<F, MAX_N_GF_TIMES>);

// Convenience bundle.
pub(crate) struct ValGrad {
    pub(crate) value: f64,
    pub(crate) gradient: Option<f64>,
}

//--------------------------------------------------------------------------------------------------
// Minor methods.

impl<F: Num> GeneFlowTimes<F> {
    pub(crate) fn new() -> Self {
        Self(ArrayVec::new())
    }
}

impl<F: Num> Parameters<F> {
    pub(crate) fn n_gf_times(&self) -> usize {
        self.gf_times.0.len()
    }
}

// Drop gradient information
impl From<&ValGrad> for f64 {
    fn from(val: &ValGrad) -> Self {
        val.value
    }
}

//==================================================================================================
// Display.

impl<F: Num + Display> Display for GeneFlowTimes<F> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.debug_list()
            .entries(self.0.iter().map(|gf| format!("{gf}")))
            .finish()
    }
}

impl Display for ValGrad {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let Self { value, gradient } = self;
        write!(f, "{value} '")?;
        if let Some(grad) = gradient {
            write!(f, "{grad}")
        } else {
            f.write_str("<nograd>")
        }
    }
}

impl fmt::Debug for ValGrad {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{self}")
    }
}
