// Implementation of BFGS algorithm with tch-rs,
// taken from Nocedal & Wright (2006).
// Stops on its own when the mean slope of the history
// becomes flat, provided the history is full.

use std::{
    fs::File,
    io::{self, Write},
    path::PathBuf,
};

use serde::Serialize;
use snafu::{ensure, ResultExt, Snafu};
use tch::Tensor;

use crate::optim::{
    simple_optim_result_impl,
    wolfe_search::{
        self, Config as WolfeSearchConfig, Error as WolfeSearchError, LastStep, Location,
        Summary as WolfeSearchSummary,
    },
    Best, Error as OptimError, Optim, OptimTensor, SlopeTracker, SlopeTrackingConfig,
};

// The exposed config.
#[derive(Debug, Serialize)]
pub(crate) struct Config {
    pub(crate) max_iter: u64,
    pub(crate) wolfe: WolfeSearchConfig,
    // If required.
    pub(crate) slope_tracking: Option<SlopeTrackingConfig>,
    // Logging.
    pub(crate) main_trace_path: Option<PathBuf>,
    pub(crate) linsearch_trace_path: Option<PathBuf>,
    // When gradient from one step to the next happen to be equal,
    // the hessian approximation breaks with a division by zero.
    // In this situation, consider that the search is over
    // if the step norm was shorter than this threshold value.
    pub(crate) small_step: f64,
    // Paths for recording optimisation details.
}

#[derive(Debug)]
pub(crate) struct BfgsResult {
    loss: f64,
    vars: Tensor,
    n_eval: u64,
    n_diff: u64,
}
simple_optim_result_impl!(BfgsResult);

// The actual search state.
struct BfgsSearch<'c, F: Fn(&Tensor) -> Tensor> {
    // Search configuration.
    cf: &'c Config,

    // Objective function,
    fun: F,
    n_vars: usize, // Cached: expensive to query.

    // Current variables values.
    vars: Tensor,
    // Current loss.
    loss: f64,
    // Current grad.
    grad: Tensor,

    // Running estimate of the *inverse* hessian matrix.
    hess: Tensor,
    // (initial value + recycled constant)
    eye: Tensor,

    // Current search direction.
    dir: Tensor,

    // Iteration number.
    n_steps: u64,
    // Keep track of total number of function evaluation and differentiation.
    n_eval: u64,
    n_diff: u64,

    // Keep track of the best point found so far.
    best: Best,

    // If desired.
    slope_tracker: Option<SlopeTracker<'c>>,

    // Log files to write to.
    main_file: Option<File>, // Main search steps.
    lin_file: Option<File>,  // Linear search on every step.
}

impl Optim for Config {
    type Error = Error;
    type Result = BfgsResult;
    fn minimize(
        &self,
        fun: impl Fn(&Tensor) -> Tensor,
        init: &Tensor,
        log: u64,
    ) -> Result<Self::Result, OptimError> {
        let n_parms = init.size1().unwrap_or_else(|e| {
            panic!(
                "Parameters must be single-dimensional. Received dimensions {:?} instead:\n{e}",
                init.size(),
            )
        });
        let (kind, device) = (init.kind(), init.device());

        // Evaluate initial gradient value.
        let mut vars = init.set_requires_grad(true);
        vars.zero_grad();
        let loss_t = fun(&vars);
        let loss = loss_t.to_double();
        ensure!(loss.is_finite(), NonFiniteLossErr { loss, vars });

        loss_t.backward();
        let grad = vars.grad().copy();
        ensure!(grad.is_all_finite(), NonFiniteGradErr { grad, vars, loss });

        let best = Best::new(loss, &vars);

        // Initial hessian estimate is plain identity.
        let eye = Tensor::eye(n_parms, (kind, device));
        let hess = eye.copy();

        // First direction to search.
        let dir = (-&hess).mv(&grad);

        // Will only be differentiated as `x + alpha * p` during linear searches.
        vars = vars.detach();

        // Initialize slope tracking if desired.
        let slope_tracker = self.slope_tracking.as_ref().map(|s| {
            let mut t = SlopeTracker::new(s, (kind, device));
            assert!(t.low_slope(loss, 0).is_none());
            t
        });

        let create_file = |path: &Option<PathBuf>| -> Result<_, Error> {
            Ok(if let Some(path) = path {
                Some(File::create(path).context(TraceErr { path })?)
            } else {
                ensure!(log == 0, NoTracePathErr { log });
                None
            })
        };

        // Spin actual search steps.
        let mut search = BfgsSearch {
            fun,
            n_vars: vars.size1().unwrap().try_into().unwrap(),
            vars,
            loss,
            grad,
            hess,
            eye,
            dir,
            n_steps: 0,
            n_eval: 1,
            n_diff: 1,
            main_file: create_file(&self.main_trace_path)?,
            lin_file: create_file(&self.linsearch_trace_path)?,
            cf: self,
            slope_tracker,
            best,
        };

        Ok(search.run_search(log)?)
    }
}

impl<F: Fn(&Tensor) -> Tensor> BfgsSearch<'_, F> {
    fn run_search(&mut self, log: u64) -> Result<BfgsResult, Error> {
        let cf = self.cf;
        if cf.max_iter == 0 {
            println!("No optimisation step asked for.");
            return Ok(self.result());
        }

        if log > 0 {
            self.write_header()?;
            wolfe_search::write_header(self.lin_file.as_mut().unwrap(), self.n_vars)?;
            self.log(0.)?; // Fake initial zero step.
        }

        loop {
            let log_this_one = log > 0 && self.n_steps % log == 0;

            // Pick a search direction.
            self.dir = (-&self.hess).mv(&self.grad);

            // Linear-search for an acceptable step size.
            let WolfeSearchSummary { last_step: res, n_eval, n_diff } = wolfe_search::search(
                Location::new(&self.fun, &self.dir, &self.vars, self.loss, &self.grad),
                &self.cf.wolfe,
                &mut self.best,
                (
                    if log_this_one { log } else { 0 },
                    self.lin_file.as_mut(),
                    self.n_steps,
                ),
            )?;
            self.n_eval += n_eval;
            self.n_diff += n_diff;
            let &mut Self { n_steps, n_eval, n_diff, .. } = self;
            let Some(LastStep {
                step_size,
                step,
                vars_after_step,
                loss_after_step,
                grad_after_step,
            }) = res
            else {
                println!(
                    "Reached null step size after {n_steps} steps: \
                     {n_eval} evaluations and {n_diff} differentiations."
                );
                break Ok(self.result());
            };

            // Check step norm.
            let norm = step.dot(&step).to_double();
            if norm <= 0.0 {
                println!(
                    "Reach silent step after {n_steps} steps: \
                     {n_eval} evaluations and {n_diff} differentiations."
                );
                break Ok(self.result());
            }
            if norm < self.cf.small_step && self.grad == grad_after_step {
                println!(
                    "Reach silent grad step after {n_steps} steps: \
                     {n_eval} evaluations and {n_diff} differentiations."
                );
                break Ok(self.result());
            }

            // Accept step.
            self.vars = vars_after_step;
            self.n_steps += 1;
            let n_steps = self.n_steps;

            // Check slope.
            if let Some(ref mut tracker) = self.slope_tracker {
                if let Some(lowslope) = tracker.low_slope(self.loss, n_steps) {
                    println!("Weak loss slope ({lowslope:e}) on iteration {n_steps}: stopping.");
                    break Ok(self.result());
                }
            }

            // Update hessian approximation (eq. 6.17).
            let Self { ref grad, ref hess, .. } = *self;
            let diff = &grad_after_step - grad;
            let rho = 1. / &diff.dot(&step);
            let left = &self.eye - &rho * step.outer(&diff);
            let right = &self.eye - &rho * diff.outer(&step);
            self.hess = left.mm(&hess.mm(&right)) + rho * step.outer(&step);

            self.loss = loss_after_step;
            self.grad = grad_after_step;

            if log_this_one {
                self.log(step_size)?;
            }

            if n_steps >= cf.max_iter {
                println!("Max iteration reached: {n_steps}.");
                break Ok(self.result());
            }
        }
    }

    fn result(&self) -> BfgsResult {
        let &Self { ref best, n_eval, n_diff, .. } = self;
        BfgsResult {
            vars: best.vars.detach().copy(),
            loss: best.loss,
            n_eval,
            n_diff,
        }
    }

    // Traces.
    fn write_header(&mut self) -> Result<(), Error> {
        // Main headers.
        let mut header = ["step", "loss", "step_size"]
            .into_iter()
            .map(ToString::to_string)
            .collect::<Vec<_>>();
        for vec in ["vars", "grad", "dir"] {
            for i in 0..self.n_vars {
                header.push(format!("{vec}_{i}"));
            }
        }
        let file = self.main_file.as_mut().unwrap();
        let path = self.cf.main_trace_path.as_ref().unwrap();
        writeln!(file, "{}", header.join(",")).with_context(|_| TraceErr { path })?;

        Ok(())
    }

    fn log(&mut self, step_size: f64) -> Result<(), Error> {
        macro_rules! w {
            ($fmt:literal $(, $val:expr)?) => {{
                let file = self.main_file.as_mut().unwrap();
                let path = self.cf.main_trace_path.as_ref().unwrap();
                write!(file, $fmt $(, $val)?).with_context(|_| TraceErr{path})
            }};
        }
        w!("{}", self.n_steps)?;
        w!(",{}", self.loss)?;
        w!(",{}", step_size)?;
        for t in [&self.vars, &self.grad, &self.dir] {
            for f in t.iter::<f64>().unwrap() {
                w!(",{f}")?;
            }
        }
        w!("\n")
    }
}

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(display("Obtained non-finite loss ({loss}) with these variables:\n{vars:?}"))]
    NonFiniteLoss { loss: f64, vars: Tensor },
    #[snafu(display(
        "Obtained non-finite gradient for loss ({loss}):\ngradient: {grad:?}\nvariables: {vars:?}"
    ))]
    NonFiniteGrad {
        grad: Tensor,
        vars: Tensor,
        loss: f64,
    },
    #[snafu(transparent)]
    WolfeSearch { source: WolfeSearchError },
    #[snafu(display("Error writing to trace file {:?}:\n{source}", path.display()))]
    Trace { path: PathBuf, source: io::Error },
    #[snafu(display(
        "A log level of {log} is required, \
         but no file path has been choosen to record traces."
    ))]
    NoTracePath { log: u64 },
}
